# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

from io import BytesIO

from PIL import Image

from pixano.features.utils.image import image_to_binary


def image_to_thumbnail(image: bytes | Image.Image) -> bytes:
    """Generate image thumbnail.

    Args:
        image: Image as binary or as a Pillow Image.

    Returns:
        Image thumbnail as binary.
    """
    if isinstance(image, bytes):
        image = Image.open(BytesIO(image))

    image.thumbnail((128, 128))
    return image_to_binary(image)



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

from pydantic import ConfigDict, field_validator
from typing_extensions import TypeVar

from pixano.app.models.table_info import TableInfo
from pixano.features import Source
from pixano.features.schemas.schema_group import SchemaGroup

from .base_schema import BaseSchemaModel


T = TypeVar("T", bound=Source)


class SourceModel(BaseSchemaModel[Source]):
    """Model for the [Source][pixano.features.Source] schema."""

    model_config = ConfigDict(
        validate_assignment=True,
        json_schema_extra={
            "examples": [
                {
                    "id": "source_0",
                    "table_info": {"group": "source", "name": "source", "base_schema": "Source"},
                    "data": {
                        "name": "source_0",
                        "kind": "model",
                        "metadata": '\\{"model_id": "model_0"\\}',
                    },
                }
            ]
        },
    )

    @field_validator("table_info")
    @classmethod
    def _validate_table_info(cls, value: TableInfo) -> TableInfo:
        """Validate table info."""
        if value.group != SchemaGroup.SOURCE.value:
            raise ValueError(f"Table info group must be {SchemaGroup.SOURCE.value}.")
        return value

    def to_row(self, schema_type: type[T]) -> T:
        """Create a [Source][pixano.features.Source] from the model."""
        if not issubclass(schema_type, Source):
            raise ValueError(f"Schema type must be a subclass of {Source.__name__}.")
        return super().to_row(schema_type)



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

import base64
import io
from pathlib import Path
from typing import Literal, overload
from urllib.error import URLError
from urllib.parse import urlparse
from urllib.request import urlopen

import PIL
from PIL import Image as PILImage

from pixano.utils import issubclass_strict

from ...types.schema_reference import ItemRef, ViewRef
from ..registry import _register_schema_internal
from .view import View


@_register_schema_internal
class Image(View):
    """Image view.

    Attributes:
        url: The image URL. Can be relative or absolute or a data URL.
        width: The image width.
        height: The image height.
        format: The image format.
    """

    url: str
    width: int
    height: int
    format: str

    @overload
    def open(self, media_dir: Path, output_type: Literal["base64"]) -> str: ...
    @overload
    def open(self, media_dir: Path, output_type: Literal["image"]) -> PILImage.Image: ...
    def open(self, media_dir: Path, output_type: Literal["base64", "image"] = "base64") -> str | PILImage.Image:
        """Open the image.

        Args:
            media_dir: Path to the media directory. If the URL is relative, it is relative to this directory.
            output_type: The output type. Can be "base64" or "image" (PIL.Image).

        Returns:
            opened image.
        """
        return Image.open_url(self.url, media_dir, output_type)

    @staticmethod
    def open_url(
        url: str, media_dir: Path, output_type: Literal["base64", "image"] = "base64"
    ) -> str | PILImage.Image:
        """Open an image from a URL.

        Args:
            url: image url relative to media_dir or absolute.
            media_dir: path to the media directory.
            output_type: output type.

        Returns:
            The opened image.
        """
        # URI is incomplete
        if urlparse(url).scheme == "":
            uri_prefix = media_dir.absolute().as_uri()
            # URI prefix exists
            if uri_prefix is not None:
                parsed_uri = urlparse(uri_prefix)
                # URI prefix is incomplete
                if parsed_uri.scheme == "":
                    raise ValueError("URI prefix is incomplete, " "no scheme provided (http://, file://, ...)")
                if url.startswith("/"):
                    url = url[1:]
                combined_path = Path(parsed_uri.path) / url
                parsed_uri = parsed_uri._replace(path=str(combined_path))
                api_url = parsed_uri.geturl()
            else:
                # No URI prefix
                raise ValueError("URI is incomplete, need URI prefix")
        # URI is already complete
        else:
            api_url = url

        try:
            with urlopen(api_url) as f:
                im_bytes = f.read()
        except URLError:
            # TODO: raise error
            print(f"Error: image not found ({api_url})")
            return ""

        if im_bytes is not None:
            if output_type == "base64":
                encoded = base64.b64encode(im_bytes).decode("utf-8")
                return f"data:image;base64,{encoded}"
            elif output_type == "image":
                return PILImage.open(io.BytesIO(im_bytes))
        # TODO: raise error
        print("Error: Image not found")
        return ""


def is_image(cls: type, strict: bool = False) -> bool:
    """Check if the given class is `Image` or a subclass of `Image`."""
    return issubclass_strict(cls, Image, strict)


def create_image(
    url: Path,
    id: str = "",
    item_ref: ItemRef = ItemRef.none(),
    parent_ref: ViewRef = ViewRef.none(),
    width: int | None = None,
    height: int | None = None,
    format: str | None = None,
    url_relative_path: Path | None = None,
) -> Image:
    """Create an `Image` instance.

    Args:
        url: The image URL. If not relative, the URL is converted to a relative path using `other_path`.
        id: Image ID.
        item_ref: Item reference.
        parent_ref: Parent view reference.
        width: The image width. If None, the width is extracted from the image file.
        height: The image height. If None, the height is extracted from the image file.
        format: The image format. If None, the format is extracted from the image file.
        url_relative_path: The path to convert the URL to a relative path.

    Returns:
        The created `Image` instance.
    """
    none_conditions = [width is None, height is None, format is None]
    not_none_conditions = [width is not None, height is not None, format is not None]
    if not all(none_conditions) and not all(not_none_conditions):
        raise ValueError("width, height and format must be all defined or all None")

    url = Path(url)

    if width is None:
        img = PIL.Image.open(url)
        width = img.width
        height = img.height
        format = img.format

    if url_relative_path is not None:
        url_relative_path = Path(url_relative_path)
        url = url.relative_to(url_relative_path)

    return Image(
        id=id, item_ref=item_ref, parent_ref=parent_ref, url=url.as_posix(), width=width, height=height, format=format
    )



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

from pydantic import ConfigDict, field_validator
from typing_extensions import TypeVar

from pixano.app.models.table_info import TableInfo
from pixano.features import Annotation
from pixano.features.schemas.schema_group import SchemaGroup

from .base_schema import BaseSchemaModel


T = TypeVar("T", bound=Annotation)


class AnnotationModel(BaseSchemaModel[Annotation]):
    """Model for the [Annotation][pixano.features.Annotation] schema."""

    model_config = ConfigDict(
        validate_assignment=True,
        json_schema_extra={
            "examples": [
                {
                    "id": "bbox_orange_cat_n1",
                    "table_info": {"group": "annotations", "name": "bboxes", "base_schema": "BBox"},
                    "data": {
                        "item_ref": {"name": "item", "id": "1"},
                        "view_ref": {"name": "image", "id": "orange_cats"},
                        "entity_ref": {
                            "name": "cats",
                            "id": "cat_n1",
                        },
                        "coords": [0, 0, 2, 2],
                        "format": "xywh",
                        "is_normalized": False,
                        "confidence": 0.8,
                    },
                }
            ]
        },
    )

    @field_validator("table_info")
    @classmethod
    def _validate_table_info(cls, value: TableInfo) -> TableInfo:
        """Validate table info."""
        if value.group != SchemaGroup.ANNOTATION.value:
            raise ValueError(f"Table info group must be {SchemaGroup.ANNOTATION.value}.")
        return value

    def to_row(self, schema_type: type[T]) -> T:
        """Create an [Annotation][pixano.features.Annotation] from the model."""
        if not issubclass(schema_type, Annotation):
            raise ValueError(f"Schema type must be a subclass of {Annotation.__name__}.")
        return super().to_row(schema_type)



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

from pydantic import ConfigDict, field_validator
from typing_extensions import TypeVar

from pixano.app.models.table_info import TableInfo
from pixano.features import View
from pixano.features.schemas.schema_group import SchemaGroup

from .base_schema import BaseSchemaModel


T = TypeVar("T", bound=View)


class ViewModel(BaseSchemaModel[View]):
    """Model for the [View][pixano.features.View] schema."""

    model_config = ConfigDict(
        validate_assignment=True,
        json_schema_extra={
            "examples": [
                {
                    "id": "orange_cats",
                    "table_info": {"group": "views", "name": "image", "base_schema": "Image"},
                    "data": {
                        "item_ref": {"name": "item", "id": "1"},
                        "parent_ref": {"name": "", "id": ""},
                        "url": "/path/to/chat/orange_left.jpg",
                        "format": "JPEG",
                        "width": 100,
                        "height": 100,
                    },
                }
            ]
        },
    )

    @field_validator("table_info")
    @classmethod
    def _validate_table_info(cls, value: TableInfo) -> TableInfo:
        """Validate table info."""
        if value.group != SchemaGroup.VIEW.value:
            raise ValueError(f"Table info group must be {SchemaGroup.VIEW.value}.")
        return value

    def to_row(self, schema_type: type[T]) -> T:
        """Create a [View][pixano.features.View] from the model."""
        if not issubclass(schema_type, View):
            raise ValueError(f"Schema type must be a subclass of {View.__name__}.")
        return super().to_row(schema_type)



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

import numpy as np
from pydantic import model_validator

from pixano.utils import issubclass_strict

from .base_type import BaseType
from .registry import _register_type_internal


@_register_type_internal
class NDArrayFloat(BaseType):
    """Represents an N-dimensional array of floating-point values.

    Attributes:
        values: The list of floating-point values in the array.
        shape: The shape of the array, represented as a list of integers.
    """

    values: list[float]
    shape: list[int]

    @model_validator(mode="after")
    def _validate_fields(self):
        if len(self.shape) < 1:
            raise ValueError("Shape must have at least one element.")
        elif any(s < 1 for s in self.shape):
            raise ValueError("Shape elements must be positive.")
        elif len(self.values) != np.prod(self.shape):
            raise ValueError("Number of values must match the product of the shape.")
        return self

    @classmethod
    def none(cls) -> "NDArrayFloat":
        """Utility function to get a `None` equivalent.
        Should be removed as soon as Lance manages `None` value.

        Returns:
            "None" `NDArrayFloat`.
        """
        return cls(values=[0], shape=[1])

    @staticmethod
    def from_numpy(arr: np.ndarray) -> "NDArrayFloat":
        """Create an instance of the class from a NumPy array.

        Args:
            arr: The NumPy array to convert.

        Returns:
            An instance of the class with values and shape derived from the input array.
        """
        return NDArrayFloat(values=arr.reshape(-1).tolist(), shape=list(arr.shape))


def is_ndarray_float(cls: type, strict: bool = False) -> bool:
    """Check if a class is a `NDArrayFloat` or a subclass of `NDArrayFloat`."""
    return issubclass_strict(cls, NDArrayFloat, strict)


def create_ndarray_float(values: list[float], shape: list[int]) -> NDArrayFloat:
    """Create a `NDArrayFloat` instance.

    Args:
        values: The list of floating-point values in the array.
        shape: The shape of the array, represented as a list of integers.

    Returns:
        The created `NDArrayFloat` instance.
    """
    return NDArrayFloat(values=values, shape=shape)



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

import warnings

from pydantic import model_validator
from typing_extensions import Self

from pixano.utils import issubclass_strict

from ...types.schema_reference import AnnotationRef, EntityRef, ItemRef, SourceRef, ViewRef
from ..registry import _register_schema_internal
from .annotation import Annotation


@_register_schema_internal
class NamedEntity(Annotation):
    """Observation of a named-entity in a text.

    Attributes:
        concept_id: Id of the named-entity type.
        mention: Named-entity observed mention.
        spans_start: List of start offsets of the spans in the text.
        spans_end: List of end offsets of the spans in the text.
    """

    concept_id: str
    mention: str
    spans_start: list[int]
    spans_end: list[int]

    @model_validator(mode="after")
    def _validate_fields(self) -> Self:
        if len(self.spans_start) != len(self.spans_end):
            raise ValueError("Spans offset lists should have the same length")
        if len(self.spans_start) == 0:
            warnings.warn("To ground a NamedEntity in a text, spans offsets should not be empty", category=UserWarning)
        if not all(span[0] >= 0 and span[1] >= 0 for span in zip(self.spans_start, self.spans_end)):
            raise ValueError("Spans offset must be positive")
        if not all(span[0] <= span[1] for span in zip(self.spans_start, self.spans_end)):
            raise ValueError("End offsets must be greater or equal to their corresponding start offset")
        return self

    @classmethod
    def none(cls) -> Self:
        """Utility function to get a `None` equivalent.
        Should be removed as soon as Lance manages `None` value.

        Returns:
            "None" NamedEntity.
        """
        return cls(
            id="",
            item=ItemRef.none(),
            view=ViewRef.none(),
            entity=EntityRef.none(),
            concept_id="",
            mention="",
            spans_start=[],
            spans_end=[],
        )

    @property
    def spans(self) -> list[tuple[int, int]]:
        """Get the list of zipped spans offsets (starts and ends)."""
        return list(zip(self.spans_start, self.spans_end))

    @property
    def spans_length(self) -> list[int]:
        """Get the computed list of spans lengths."""
        return [e - s for s, e in zip(self.spans_start, self.spans_end)]


@_register_schema_internal
class Relation(Annotation):
    """Observation of a relation between two named entity in a text.

    Attributes:
        predicate_id: id of relation type.
        subject_id: annotation_id of the subject named-entity
        object_id: annotation_id of the object named-entity
    """

    predicate_id: str
    subject_id: AnnotationRef
    object_id: AnnotationRef

    @classmethod
    def none(cls) -> Self:
        """Utility function to get a `None` equivalent.
        Should be removed as soon as Lance manages `None` value.

        Returns:
            "None" Relation.
        """
        return cls(
            id="",
            item=ItemRef.none(),
            view=ViewRef.none(),
            entity=EntityRef.none(),
            predicate_id="",
            subject_id=AnnotationRef.none(),
            object_id=AnnotationRef.none(),
        )


def is_named_entity(cls: type, strict: bool = False) -> bool:
    """Check if a class is a `NamedEntity` or subclass of `NamedEntity`."""
    return issubclass_strict(cls, NamedEntity, strict)


def is_relation(cls: type, strict: bool = False) -> bool:
    """Check if a class is a `Relation` or subclass of `Relation`."""
    return issubclass_strict(cls, Relation, strict)


def create_named_entity(
    concept_id: str,
    mention: str,
    spans_start: list[int],
    spans_end: list[int],
    id: str = "",
    item_ref: ItemRef = ItemRef.none(),
    view_ref: ViewRef = ViewRef.none(),
    entity_ref: EntityRef = EntityRef.none(),
    source_ref: SourceRef = SourceRef.none(),
) -> NamedEntity:
    """Create a `NamedEntity` instance.

    Args:
        concept_id: Id of named-entity type.
        mention: Named-entity observed mention.
        spans_start: List of start offsets of the spans in the text.
        spans_end: List of end offsets of the spans in the text.
        id: `NamedEntity` ID.
        item_ref: Item reference.
        view_ref: View reference.
        entity_ref: Entity reference.
        source_ref: Source reference.

    Returns:
        The created `NamedEntity` instance.
    """
    return NamedEntity(
        concept_id=concept_id,
        mention=mention,
        spans_start=spans_start,
        spans_end=spans_end,
        id=id,
        item_ref=item_ref,
        view_ref=view_ref,
        entity_ref=entity_ref,
        source_ref=source_ref,
    )


def create_relation(
    predicate_id: str,
    subject_id: AnnotationRef = AnnotationRef.none(),
    object_id: AnnotationRef = AnnotationRef.none(),
    id: str = "",
    item_ref: ItemRef = ItemRef.none(),
    view_ref: ViewRef = ViewRef.none(),
    entity_ref: EntityRef = EntityRef.none(),
    source_ref: SourceRef = SourceRef.none(),
) -> Relation:
    """Create a `Relation` instance.

    Args:
        predicate_id: Id of relation type.
        subject_id: Annotation_id of the subject named-entity
        object_id: Annotation_id of the object named-entity
        id: `Relation` ID.
        item_ref: Item reference.
        view_ref: View reference.
        entity_ref: Entity reference.
        source_ref: Source reference.

    Returns:
        The created `Relation` instance.
    """
    return Relation(
        predicate_id=predicate_id,
        subject_id=subject_id,
        object_id=object_id,
        id=id,
        item_ref=item_ref,
        view_ref=view_ref,
        entity_ref=entity_ref,
        source_ref=source_ref,
    )



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

from enum import Enum

from pixano.features.schemas.source import Source

from .annotations import Annotation
from .embeddings import Embedding
from .entities import Entity
from .items import Item
from .views import View


class SchemaGroup(Enum):
    """Schema group.

    It defines the different schema groups to which a schema can belong.

    Attributes:
        ANNOTATION: Annotation schema group.
        EMBEDDING: Embedding schema group.
        ITEM: Item schema group.
        ENTITY: Entity schema group.
        VIEW: View schema group.
    """

    ANNOTATION = "annotations"
    EMBEDDING = "embeddings"
    ITEM = "item"
    ENTITY = "entities"
    VIEW = "views"
    SOURCE = "source"

    @classmethod
    def _missing_(cls, name):
        for member in cls:
            if member.value == name.lower():
                return member


_SCHEMA_GROUP_TO_SCHEMA_DICT = {
    SchemaGroup.EMBEDDING: Embedding,
    SchemaGroup.ITEM: Item,
    SchemaGroup.ENTITY: Entity,
    SchemaGroup.ANNOTATION: Annotation,
    SchemaGroup.VIEW: View,
    SchemaGroup.SOURCE: Source,
}
